/* 
 * minlabel.c
 *
 * Copyright 1995 Digital Equipment Corporation, Litleton, MA
 *
 * This file may be redistributed according to the terms of the
 * GNU General Public License.
 */

/* Hacked by Erik Troan (ewt@redhat.com) to add --version flag and
   return rc=1 on bad useage */

#include <strings.h>
#include <sys/types.h>
#include <linux/hdreg.h>
#define DKTYPENAMES
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <linux/kdev_t.h>

#define MAXPARTITIONS	(8)
#define DISKLABELMAGIC	(0x82564557UL)

#define VERSION "1.0"

struct disklabel {
	unsigned int d_magic;
	unsigned short d_type,d_subtype;
	unsigned char d_typename[16];
	unsigned char d_packname[16];
	unsigned int d_secsize;
	unsigned int d_nsectors;
	unsigned int d_ntracks;
	unsigned int d_ncylinders;
	unsigned int d_secpercyl;
	unsigned int d_secperunit;
	unsigned short d_sparespertrack;
	unsigned short d_sparespercyl;
	unsigned int d_acylinders;
	unsigned short d_rpm, d_interleave, d_trackskew, d_cylskew;
	unsigned int d_headswitch, d_trkseek, d_flags;
	unsigned int d_drivedata[5];
	unsigned int d_spare[5];
	unsigned int d_magic2;
	unsigned short d_checksum;
	unsigned short d_npartitions;
	unsigned int d_bbsize, d_sbsize;
	struct d_partition {
		unsigned int p_size;
		unsigned int p_offset;
		unsigned int p_fsize;
		unsigned char  p_fstype;
		unsigned char  p_frag;
		unsigned short p_cpg;
	} d_partitions[MAXPARTITIONS];
};

static char *fstypenames[] = {
        "unused",	/* 0 */
        "swap",		/* 1 */
        "Version 6",	/* 2 */
        "Version 7",	/* 3 */
        "System V",	/* 4 */
        "4.1BSD",	/* 5 */
        "Eighth Edition", /* 6 */
        "4.2BSD",	/* 7 */
        "ext2",		/* 8 */
        "boot",		/* 9 */
        "resrvd10",	/* 10 */
        "resrvd11",	/* 11 */
        "resrvd12",	/* 12 */
        "resrvd13",	/* 13 */
        "resrvd14",	/* 14 */
        "resrvd15",	/* 15 */
        "ADVfs",	/* 16 */
        "LSMpubl",	/* 17 */
        "LSMpriv",	/* 18 */
        "LSMsimp",	/* 19 */
        0
};

#define FSMAXTYPES 20

static char *dktypenames[] = {
        "unknown",
        "SMD",
        "MSCP",
        "old DEC",
        "SCSI",
        "ESDI",
        "type 6",
        "type 7",
        "type 8",
        "type 9",
        "floppy",
        "SWXCR",
        0
};

#define DKMAXTYPES 12

#define         D_REMOVABLE     0x01            /* removable media */
#define         D_ECC           0x02            /* supports ECC */
#define         D_BADSECT       0x04            /* supports bad sector forw. */
#define         D_RAMDISK       0x08            /* disk emulator */
#define         D_CHAIN         0x10            /* can do back-back transfers */


char	typename[16];
char	packname[16];
int readonly = 0;

typedef union __bootblock {
    struct {
        char			__pad1[64];
        struct disklabel	__label;
    } __u1;
    struct {
	unsigned long		__pad2[63];
	unsigned long		__checksum;
    } __u2;
    char		bootblock_bytes[512];
    unsigned short	bootblock_shorts[256];
    unsigned int	bootblock_longs[128];
    unsigned long	bootblock_quadwords[64];
} bootblock;

#define	bootblock_label		__u1.__label
#define bootblock_checksum	__u2.__checksum

static int	bootblock_modified = 0;

read_bootblock(int fd, bootblock *bb)
{
    lseek(fd, 0L, 0);
    read(fd, bb, sizeof(bootblock));
}

write_bootblock(int fd, bootblock *bb)
{
    if(readonly) {
	fprintf(stderr, 
		"Utility invoked in readonly mode; NOT writing disklabel!\n");
    }
    else {
        lseek(fd, 0L, 0);
        write(fd, bb, sizeof(bootblock));
    }
}

void initialize_label(int fd, struct disklabel *dl)
{
    /* Initialize the disklabel.  We'll fill in the geometry fields,
     * any other fields we can, and provide a suggested configuration.
     * The user is free to change this.
     */
    
    struct	hd_geometry hg;
    int		retval;

    memset(dl, 0, sizeof(struct disklabel));

    /* Magic number */
    dl->d_magic = dl->d_magic2 = DISKLABELMAGIC;
    dl->d_type = 8 /* ext2 */;
    strcpy(dl->d_typename, typename);
    strcpy(dl->d_packname, packname);

    retval = ioctl(fd, HDIO_GETGEO, &hg);
    if(retval < 0) {
        perror("HDIO_GETGEO");
        close(fd);
        exit(0);
    }

    dl->d_secsize = 512;
    dl->d_nsectors = hg.sectors;
    dl->d_ntracks =  hg.heads;
    dl->d_ncylinders = hg.cylinders;
    dl->d_secpercyl = hg.heads * hg.sectors;
    dl->d_secperunit = dl->d_secpercyl * dl->d_ncylinders;
    dl->d_rpm = 3600;
    dl->d_interleave = 1;

    dl->d_npartitions = MAXPARTITIONS;

    /* Create two partitions; the first one for booting (8Mb), the
     * rest for data.
     */
    dl->d_partitions[0].p_size = (8 * 1024 * 1024) / dl->d_secsize;
    dl->d_partitions[0].p_offset = 0;
    dl->d_partitions[0].p_fsize = 1024;
    dl->d_partitions[0].p_fstype = 9 /* boot */;
    dl->d_partitions[0].p_frag = 0;
    dl->d_partitions[0].p_cpg = 0;

    dl->d_partitions[1].p_size = dl->d_secperunit - dl->d_partitions[0].p_size;
    dl->d_partitions[1].p_offset = dl->d_partitions[0].p_size;
    dl->d_partitions[1].p_fsize = 1024;
    dl->d_partitions[1].p_fstype = 8 /* ext2 */;
    dl->d_partitions[1].p_frag = 0;
    dl->d_partitions[1].p_cpg = 0;

}


void display_label(lp)
	struct disklabel *lp;
{
	int i, j;
	struct d_partition *pp;

	fprintf(stderr, "# %s:\n", packname);
	if ((unsigned) lp->d_type < DKMAXTYPES)
		fprintf(stderr, "type: %s\n", dktypenames[lp->d_type]);
	else
		fprintf(stderr, "type: %d\n", lp->d_type);
        fprintf(stderr, "disk: %s, label: %s\n", 
				lp->d_typename, lp->d_packname);

	fprintf(stderr, "flags:");
	if (lp->d_flags & D_REMOVABLE)
		fprintf(stderr, " removable");
	if (lp->d_flags & D_ECC)
		fprintf(stderr, " ecc");
	if (lp->d_flags & D_BADSECT)
		fprintf(stderr, " badsect");
	fprintf(stderr, "\n");
	fprintf(stderr, "bytes/sector: %d    ", lp->d_secsize);
	fprintf(stderr, "sectors/track: %d\n", lp->d_nsectors);
	fprintf(stderr, "tracks/cylinder: %d    ", lp->d_ntracks);
	fprintf(stderr, "sectors/cylinder: %d\n", lp->d_secpercyl);
	fprintf(stderr, "cylinders: %d    ", lp->d_ncylinders);
	fprintf(stderr, "sectors/unit: %d\n", lp->d_secperunit);
	fprintf(stderr, "rpm: %d    ", lp->d_rpm);
	fprintf(stderr, "interleave: %d\n", lp->d_interleave);
	fprintf(stderr, "trackskew: %d    ", lp->d_trackskew);
	fprintf(stderr, "cylinderskew: %d\n", lp->d_cylskew);
	fprintf(stderr, "headswitch: %d\t\t# milliseconds\n", lp->d_headswitch);
	fprintf(stderr, "track-to-track seek: %d\t# milliseconds\n", lp->d_trkseek);
	fprintf(stderr, "drivedata: ");
	for (i = 4; i >= 0; i--)
		if (lp->d_drivedata[i])
			break;
	if (i < 0)
		i = 0;
	for (j = 0; j <= i; j++)
		fprintf(stderr, "%d ", lp->d_drivedata[j]);
	fprintf(stderr, "\n\n%d partitions:\n", lp->d_npartitions);
	fprintf(stderr,
	    "#        size   offset    fstype   Size(Kb)   Size(Mb)\n");
	pp = lp->d_partitions;
	for (i = 0; i < lp->d_npartitions; i++, pp++) {
		int		kbytes, mbytes;

		fprintf(stderr, "  %c: %8d %8d  ", '1' + i,
		   pp->p_size, pp->p_offset);
		if ((unsigned) pp->p_fstype < FSMAXTYPES)
			fprintf(stderr, "%8.8s", fstypenames[pp->p_fstype]);
		else
			fprintf(stderr, "%8d", pp->p_fstype);

		kbytes = pp->p_size * lp->d_secsize / 1024;
		mbytes = kbytes / 1024;
		fprintf(stderr, " %10d  %6d.%1d", kbytes,
			mbytes, ((kbytes-(mbytes*1024))*10)/1024);

		fprintf(stderr, "\t# (Cyl. %4d",
		    pp->p_offset / lp->d_secpercyl);
		if (pp->p_offset % lp->d_secpercyl)
		    putc('*', stderr);
		else
		    putc(' ', stderr);
		fprintf(stderr, "- %d",
		    (pp->p_offset + 
		    pp->p_size + lp->d_secpercyl - 1) /
		    lp->d_secpercyl - 1);
		if (pp->p_size % lp->d_secpercyl)
		    putc('*', stderr);
		fprintf(stderr, ")\n");
	}
	fflush(stderr);
}

void zero_partition(struct disklabel *dl, int p)
{
    memset(&(dl->d_partitions[p]), 0, sizeof(struct d_partition));
}

int change_partition(struct disklabel *dl, int p)
{
    char	answer[256];
    int		newsize, newoffset, newtype;
    int		i;

    /* Set defaults */
    newsize = dl->d_partitions[p].p_size;
    newoffset = dl->d_partitions[p].p_offset;
    newtype = dl->d_partitions[p].p_fstype;

    if (newoffset == 0) newoffset = 1;

    fprintf(stderr, "Partition %d:\n    Size (sectors) [%d]: ",
		p+1, dl->d_partitions[p].p_size);
    if(fgets(answer, 256, stdin) == 0) return 0;
    
    if((*answer != '\r') && (*answer != '\n') && (*answer != '\0')) {
   	sscanf(answer, "%d", &newsize);
	if(newsize < 0 || newsize > dl->d_secperunit) {
	    fprintf(stderr, "Invalid size: must be between 0 and %d\n",
			dl->d_secperunit);
	    return 0;
	}
    }


    fprintf(stderr, "    Offset (sectors) [%d]: ", 
				dl->d_partitions[p].p_offset ? dl->d_partitions[p].p_offset : 1);
    if(fgets(answer, 256, stdin) == 0) return 0;
    if((*answer != '\r') && (*answer != '\n') && (*answer != '\0')) {
   	sscanf(answer, "%d", &newoffset);
	if(newoffset <= 0 || newoffset > dl->d_secperunit) {
	    fprintf(stderr, "Invalid offset: must be between 1 and %d\n",
			dl->d_secperunit);
	    return 0;
	}
    }

    fprintf(stderr, "\nFS type numbers:\n");
    for(i = 0; i <= FSMAXTYPES; i++) {
	fprintf(stderr, "%2d: %-18s   ", i, fstypenames[i]);
	if((i%3)==2) {
	    fprintf(stderr, "\n");
	}
    }

    fprintf(stderr, "\n    FS type [%d]: ", dl->d_partitions[p].p_fstype);
    if(fgets(answer, 256, stdin) == 0) return 0;
    if((*answer != '\r') && (*answer != '\n') && (*answer != '\0')) {
   	sscanf(answer, "%d", &newtype);
	if(newtype < 0 || newtype > FSMAXTYPES) {
	    fprintf(stderr, "Invalid FS type: must be between 0 and %d\n",
			FSMAXTYPES);
	    return 0;
	}
    }

    dl->d_partitions[p].p_size = newsize;
    dl->d_partitions[p].p_offset = newoffset;
    dl->d_partitions[p].p_fstype = newtype;
    return 1;
}


void command_loop(int fd, bootblock *bb)
{
    char	cmdline[256];
    char	answer[32];
    int		i, p;
    struct disklabel	*dl = &(bb->bootblock_label);
    int redisplay = 1;

    for(;;) {
	if(redisplay) {
	    display_label(dl);
	    redisplay = 0;
	}
	fprintf(stderr, "\nCommand: ");
	fflush(stderr);
	if(fgets(cmdline, 256, stdin) == 0) {
	    close(fd);
	    exit(0);
	}
	switch(*cmdline) {
	    case 'w':				/* Write new disklabel */
		write_bootblock(fd, bb);
		bootblock_modified = 0;
		break;
	    case 'r':				/* Re-read label from dsk */
		read_bootblock(fd, bb);
		bootblock_modified = 0;
		redisplay = 1;
		break;
	    case 'i':				/* Initialize disklabel. */
		initialize_label(fd, dl);
		bootblock_modified = 1;
		redisplay = 1;
		break;
	    case 'h':
	    case '?':
		fprintf(stderr, "Commands available:\n");
		fprintf(stderr, "  l - list partition table\n");
		fprintf(stderr, "  w - write label to disk (use carefully!)\n");
		fprintf(stderr, "  r - re-read label from disk\n");
		fprintf(stderr, "  i - Initialize a default label\n");
		fprintf(stderr, "  c - Change a partition\n");
		fprintf(stderr, "  a - Add a partition\n");
		fprintf(stderr, "  z - Zero a partition\n");
		fprintf(stderr, "  d - Delete a partition\n");
		fprintf(stderr, "  q - quit command loop\n");
		fprintf(stderr, "  h - print this text\n");
		break;
	    case 'l':
		redisplay = 1;
		break;
	    case 'c':
		fprintf(stderr, "Partition to change [1-%d]: ",
			dl->d_npartitions);
		fflush(stderr);
		fgets(answer, 32, stdin);
		if(*answer == '\r' || *answer == '\n' || *answer == '\0') break;
		sscanf(answer, "%d", &p);
		if((p < 1) || (p > dl->d_npartitions)) {
		    fprintf(stderr, "Invalid partition number %d\n", p);
		    break;
		}
		p--;
		if(change_partition(dl, p)) {
			bootblock_modified = 1;
			redisplay = 1;
		}
		break;
	    case 'a':
		if(dl->d_npartitions == MAXPARTITIONS) {
		    fprintf(stderr, "No more then %d partitions allowed\n",
				MAXPARTITIONS);
		    fprintf(stderr, "You must delete a partition first\n");
		    fprintf(stderr, "or modify an existing partition\n");
		}
		else {
		    p = dl->d_npartitions++;
		    if(change_partition(dl, p)) {
		        redisplay = 1;
			bootblock_modified = 1;
		    }
		}
		break;
	    case 'z':
		fprintf(stderr, "Partition to zero [0-%d]: ",
			dl->d_npartitions);
		fflush(stderr);
		fgets(answer, 32, stdin);
		if(*answer == '\r' || *answer == '\n' || *answer == '\0') break;
		sscanf(answer, "%d", &p);
		if((p < 1) || (p > dl->d_npartitions)) {
		    fprintf(stderr, "Invalid partition number %d\n", p);
		    break;
		}
		p--;
		zero_partition(dl, p);
		bootblock_modified = 1;
		redisplay = 1;
		break;
	    case 'd':
		fprintf(stderr, "Partition to delete [0-%d]: ",
			dl->d_npartitions);
		fflush(stderr);
		fgets(answer, 32, stdin);
		if(*answer == '\r' || *answer == '\n' || *answer == '\0') break;
		sscanf(answer, "%d", &p);
		if((p < 1) || (p > dl->d_npartitions)) {
		    fprintf(stderr, "Invalid partition number %d\n", p);
		    break;
		}
		p--;
		for(i = p; i < (dl->d_npartitions-1); i++) {
		    dl->d_partitions[i] = dl->d_partitions[i+1];
		}
		zero_partition(dl, dl->d_npartitions-1);
		dl->d_npartitions--;
		redisplay = 1;
		bootblock_modified = 1;
		break;
	    case 'q':
		return;
	    case ' ':
	    case '\t':
	    case '\r':
	    case '\n':
	    case '\0':
		break;
	    default:
		fprintf(stderr, "Unrecognized command: %s\n", cmdline);
		break;
	}
    }
}

int is_rawdev(dev_t d)
{
    /* Check to see if this is a "raw" device (i.e. not a partition) */
    if(MAJOR(d) == 3) {		/* hda, hdb... */
	if((MINOR(d) & 63)  == 0) {
	    return(1);
	}
	strcpy(typename, "IDE");
    }
    else if(MAJOR(d) == 8) {	/* sda, sdb... */
	if((MINOR(d) & 15) == 0) {
	    return(1);
	}
	strcpy(typename, "SCSI");
    }
    return(0);			/* no other block dev's supported... */
}


main(int argc, char **argv)
{
    int		fd;
    struct stat	st;
    bootblock	bb;
    struct disklabel * dl;
    char	answer[32];
    char	*devname;

    if(argc == 2) {
	if (!strcmp(argv[1], "--version")) {
	    printf("mklabel version %s\n", VERSION);
	    exit(0);
	}
	devname = argv[1];
    }
    else if(argc == 3) {
	if(strcmp(argv[1], "-r") == 0) {
	    readonly = 1;
	    devname = argv[2];
	}
	else {
	    fprintf(stderr, "Usage: %s [--version] [-r] device\n", argv[0]);
	    exit(1);
	}
    }
    else {
	fprintf(stderr, "Usage: %s [--version] [-r] device\n", argv[0]);
	exit(1);
    }

    /* Open the device */
    if(readonly) {
	fd = open(devname, O_RDONLY);
    }
    else {
        fd = open(devname, O_RDWR);
    }
    if(fd < 0) {
	perror(devname);
	exit(0);
    }

    /* See if the disk can be stat'ed */
    if(fstat(fd, &st) < 0) {
	perror(devname);
	exit(0);
    }

    /* See if this is the "raw" disk... */
    if(!is_rawdev(st.st_rdev)) {
	fprintf(stderr, 
		"%s is not a raw disk device.\n", devname);
	fprintf(stderr, 
		"You must specify a raw device (/dev/sda, /dev/hda),\n");
	fprintf(stderr, 
		"not a partition (/dev/sda1, /dev/hda2)\n");
	exit(0);
    }
    strcpy(packname, devname);

    /* Read in the current bootblock... */
    read_bootblock(fd, &bb);

    if(bb.bootblock_shorts[255] == 0x55aa) {
	fprintf(stderr, "WARNING: FAT-style partition table detected.\n");
	fprintf(stderr, "You must erase any FAT-style partition table\n");
	fprintf(stderr, "before establishing a disk label.  Do this now [y/n]? ");
	fflush(stderr);
	fgets(answer, 32, stdin);
	if(*answer == 'y') {
	    bb.bootblock_shorts[255] = 0;
	}
	bootblock_modified = 1;
    }

    dl = &(bb.bootblock_label);

    if((dl->d_magic != DISKLABELMAGIC) ||
	(dl->d_magic2 != DISKLABELMAGIC)) {
	fprintf(stderr, 
		"Disk currently has no label.  Initialize a new label?\n");
	fprintf(stderr, "WARNING: Initializing a new label will destory\n");
	fprintf(stderr, "all data currently on the device! [y/n]? ");
	fflush(stderr);
	fgets(answer, 32, stdin);
	if(*answer == 'y') {
	    initialize_label(fd, dl);
	    bootblock_modified = 1;
	}
	else {
	    fprintf(stderr, "Bailing out!\n");
	    close(fd);
	    exit(0);
	}
    }

    command_loop(fd, &bb);

    if(!readonly && bootblock_modified) {
	    fprintf(stderr, "Write new label to disk? [y/n] ");
	    fflush(stderr);
	    fgets(answer, 32, stdin);
	    if(*answer == 'y') {
		write_bootblock(fd, &bb);
		fprintf(stderr, "Label written\n");
	    }
	    else {
		fprintf(stderr, "Previous label remains intact\n");
	    }
    }

    close(fd);
    exit(0);
}

